import copy
import json
import uuid

from enum import Enum
from voluptuous import Schema, Invalid, MultipleInvalid, Any, ALLOW_EXTRA, PREVENT_EXTRA


class Resource(object):
    DEFAULT_ID_LENGTH = 8
    ADDITIONAL_SCHEMA = {}

    def __new__(cls, *args, **kwargs):
        attributes = cls.__attributes_from_dict(args[0] if len(args) else kwargs)
        resource_type = attributes['type'].value
        if resource_type is Resource:
            raise InvalidResourceException('Resource class is abstract')
        cls.__validate_attributes(attributes, cls.__schema(resource_type, extend=True))
        instance = super(Resource, resource_type).__new__(resource_type)
        instance._attributes = attributes  # pylint: disable=protected-access
        return instance

    @classmethod
    def __attributes_from_dict(cls, attributes):
        cls.__validate_attributes(attributes, cls.__schema(extend=False))
        attributes = copy.deepcopy(attributes)
        attributes['id'] = attributes.get('id', Resource.__generate_random_id())
        attributes['type'] = ResourceType.validate(attributes.get('type', cls))
        return attributes

    @classmethod
    def __validate_attributes(cls, attributes, schema):
        try:
            schema(attributes)
        except MultipleInvalid as invalid:
            raise InvalidResourceException(str(invalid))

    @classmethod
    def __schema(cls, resource_type=None, extend=False):
        try:
            string_validator = Any(str, unicode)
        except NameError:
            string_validator = Any(str)

        schema = Schema(schema={'id': string_validator, 'type': ResourceType.validate}, extra=ALLOW_EXTRA)
        if extend:
            additional = resource_type.ADDITIONAL_SCHEMA if resource_type else cls.ADDITIONAL_SCHEMA
            schema = schema.extend(additional, extra=PREVENT_EXTRA)
        return schema

    @property
    def type(self):
        return str(self._attributes['type'])

    def __getattr__(self, item):
        try:
            return self._attributes[item]
        except KeyError:
            raise AttributeError('Attribute (%s) is not set on resource')

    @staticmethod
    def __generate_random_id():
        return uuid.uuid4().hex[:Resource.DEFAULT_ID_LENGTH]

    def __eq__(self, other):
        try:
            for attr, other_value in other.items():
                value = self._attributes.get(attr)
                if value is None or value != other_value:
                    return False
            return True
        except AttributeError:
            return False

    def __repr__(self):
        return json.dumps(self, default=Resource.json_encoder)

    def __iter__(self):
        return self._attributes.__iter__()

    def keys(self):
        return self._attributes.keys()

    def values(self):
        return self._attributes.values()

    def items(self):
        return self._attributes.items()

    def __getitem__(self, item):
        return self._attributes[item]

    def __add__(self, other):
        return ResourceGroup(resources=[self, other])

    def flatten(self, fqn=None):
        fqn = '%s.%s' % (fqn, self.id) if fqn else self.id
        return {'%s.%s' % (fqn, key): str(value) for key, value in self.items()}

    @staticmethod
    def json_encoder(data):
        if isinstance(data, ResourceType):
            return str(data)
        return dict(data)


class ResourceGroup(Resource):
    ADDITIONAL_SCHEMA = {
        'resources': Any(list, set, tuple)
    }

    def __init__(self, *_, **__):
        self.__resources = {}
        if 'resources' in self._attributes.keys():
            for resource in self._attributes['resources']:
                self.add(resource)
        self._attributes['resources'] = self.resources

    def add(self, resource):
        if not isinstance(resource, Resource):
            try:
                resource = Resource(resource)
            except InvalidResourceException:
                raise InvalidResourceException('ResourceGroup can only store resources')
        self.__resources[resource.id] = resource

    def remove(self, resource):
        try:
            del self.__resources[Resource(resource).id]
        except InvalidResourceException:
            del self.__resources[resource.id]

    @property
    def resources(self):
        return list(self.__resources.values())

    def diff(self, other):
        flat_self = self.flatten()
        flat_other = other.flatten()
        diff_keys = set(flat_self.keys()).symmetric_difference(flat_other.keys())
        return {key: flat_self.get(key, flat_other.get(key)) for key in diff_keys}

    def flatten(self, fqn=None):
        flat = dict()
        for resource in self.resources:
            flat.update(resource.flatten('%s.%s' % (fqn, self.id) if fqn else self.id))
        return flat


class Instance(Resource):
    ADDITIONAL_SCHEMA = {
        'number_of_cores': int
    }


class ResourceType(Enum):
    null = Resource
    group = ResourceGroup
    instance = Instance

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        try:
            return cls[value]
        except KeyError:
            return cls(value)
        except ValueError:
            raise Invalid('The given resource type (%s) is not valid' % value)

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, str(self))


class InvalidResourceException(Exception):
    pass
